@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_proto_modules.c,v 1.88 2010/10/19 08:57:18 kitagawa Exp $
+ * $Header$
  * 
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2010	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -61,9 +61,6 @@ char *copy_schema = NULL;  /* copy table name */
 char copy_delimiter; /* copy delimiter char */
 char *copy_null = NULL; /* copy null string */
 
-/* non 0 if "BEGIN" query with extended query protocol received */
-int receive_extended_begin = 0;
-
 /*
  * Non 0 if allow to close internal transaction.  This variable was
  * introduced on 2008/4/3 not to close an internal transaction when
@@ -91,8 +88,7 @@ static int check_errors(POOL_CONNECTION_POOL *backend, int backend_id);
 static void generate_error_message(char *prefix, int specific_error, char *query);
 static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 									 POOL_CONNECTION_POOL *backend,
-									 PreparedStatement *ps);
-static void overwrite_map_for_deallocate(POOL_QUERY_CONTEXT *query_context);
+									 POOL_SENT_MESSAGE *message);
 static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *number_of_nodes);
 static int extract_ntuples(char *message);
 
@@ -117,6 +113,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
+	POOL_MEMORY_POOL *old_context;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -150,22 +147,40 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		return POOL_END;
 	}
 
+	/* switch memory context */
+	old_context = pool_memory;
+	pool_memory = query_context->memory_context;
+
 	/* parse SQL string */
 	parse_tree_list = raw_parser(contents);
 
 	if (parse_tree_list == NIL)
 	{
-		/*
-		 * Unable to parse the query. Probably syntax error or the
-		 * query is too new and our parser cannot understand. Treat as
-		 * if it were an DELETE command. Note that the DELETE command
-		 * does not execute, instead the original query will be sent
-		 * to backends, which may or may not cause an actual syntax errors.
-		 * The command will be sent to all backends in replication mode
-		 * or master/primary in master/slave mode.
-		 */
-		pool_log("SimpleQuery: Unable to parse the query: %s", contents);
-		parse_tree_list = raw_parser(POOL_DUMMY_QUERY);
+		/* is the query empty? */
+		if (*contents == '\0' || *contents == ';')
+		{
+			/*
+			 * JBoss sends empty queries for checking connections.
+			 * We replace the empty query with SELECT command not
+			 * to affect load balance.
+			 * [Pgpool-general] Confused about JDBC and load balancing
+			 */
+			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY);
+		}
+		else
+		{
+			/*
+			 * Unable to parse the query. Probably syntax error or the
+			 * query is too new and our parser cannot understand. Treat as
+			 * if it were an DELETE command. Note that the DELETE command
+			 * does not execute, instead the original query will be sent
+			 * to backends, which may or may not cause an actual syntax errors.
+			 * The command will be sent to all backends in replication mode
+			 * or master/primary in master/slave mode.
+			 */
+			pool_log("SimpleQuery: Unable to parse the query: %s", contents);
+			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY);
+		}
 	}
 
 	if (parse_tree_list != NIL)
@@ -216,30 +231,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		 */
 		check_copy_from_stdin(node);
 
-		/*
-		 * if this is DROP DATABASE command, send USR1 signal to parent and
-		 * ask it to close all idle connections.
-		 * XXX This is overkill. It would be better to close the idle
-		 * connection for the database which DROP DATABASE command tries
-		 * to drop. This is impossible at this point, since we have no way
-		 * to pass such info to other processes.
-		 */
-		if (is_drop_database(node))
-		{
-			int stime = 5;	/* XXX give arbitrary time to allow closing idle connections */
-
-			pool_debug("Query: sending SIGUSR1 signal to parent");
-
-			Req_info->kind = CLOSE_IDLE_REQUEST;
-			kill(getppid(), SIGUSR1);		/* send USR1 signal to parent */
-
-			/* we need to loop over here since we will get USR1 signal while sleeping */
-			while (stime > 0)
-			{
-				stime = sleep(stime);
-			}
-		}
-
 		/* status reporting? */
 		if ((IsA(node, VariableShowStmt) && strncasecmp(sq_config, contents, strlen(sq_config)) == 0)
          || (IsA(node, VariableShowStmt) && strncasecmp(sq_pools, contents, strlen(sq_pools)) == 0)
@@ -282,7 +273,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 					 sp->user, sp->database, remote_ps_data);
 			set_ps_display(psbuf, false);
 
-			free_parser();
 			pool_query_context_destroy(query_context);
 			pool_set_skip_reading_from_backends();
 			return POOL_CONTINUE;
@@ -294,6 +284,30 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		pool_where_to_send(query_context, query_context->original_query,
 						   query_context->parse_tree);
 
+		/*
+		 * if this is DROP DATABASE command, send USR1 signal to parent and
+		 * ask it to close all idle connections.
+		 * XXX This is overkill. It would be better to close the idle
+		 * connection for the database which DROP DATABASE command tries
+		 * to drop. This is impossible at this point, since we have no way
+		 * to pass such info to other processes.
+		 */
+		if (is_drop_database(node))
+		{
+			int stime = 5;	/* XXX give arbitrary time to allow closing idle connections */
+
+			pool_debug("Query: sending SIGUSR1 signal to parent");
+
+			Req_info->kind = CLOSE_IDLE_REQUEST;
+			kill(getppid(), SIGUSR1);		/* send USR1 signal to parent */
+
+			/* we need to loop over here since we will get USR1 signal while sleeping */
+			while (stime > 0)
+			{
+				stime = sleep(stime);
+			}
+		}
+
 		/*
 		 * determine if we need to lock the table
 		 * to keep SERIAL data consistency among servers
@@ -323,7 +337,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 					status = insert_lock(frontend, backend, contents, (InsertStmt *)node, lock_kind);
 					if (status != POOL_CONTINUE)
 					{
-						free_parser();
+						pool_query_context_destroy(query_context);
 						return status;
 					}
 				}
@@ -331,10 +345,11 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		}
 		else if (REPLICATION && contents == NULL && start_internal_transaction(frontend, backend, node))
 		{
-			free_parser();
+			pool_query_context_destroy(query_context);
 			return POOL_ERROR;
 		}
 	}
+
 	if (MAJOR(backend) == PROTO_MAJOR_V2 && is_start_transaction_query(node))
 	{
 		int i;
@@ -346,6 +361,24 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		}
 	}
 
+	if (node)
+	{
+		POOL_SENT_MESSAGE *msg = NULL;
+
+		if (IsA(node, PrepareStmt))
+		{
+			msg = pool_create_sent_message('Q', len, contents, 0,
+										   ((PrepareStmt *)node)->name,
+										   query_context);
+			if (!msg)
+			{
+				pool_error("SimpleQuery: cannot create query message: %s", strerror(errno));
+				return POOL_END;
+			}
+			session_context->uncompleted_message =  msg;
+		}
+	}
+
 	string = query_context->original_query;
 
 	if (!RAW_MODE)
@@ -362,20 +395,21 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 
 			if (node)
 		   	{
-				PreparedStatement *ps = NULL;
+				POOL_SENT_MESSAGE *msg = NULL;
 
 				if (IsA(node, PrepareStmt))
 				{
-#ifdef NOT_USED
-					ps = session_context->pending_pstmt;
-					ps->num_tsparams = 0;
-#endif
+					msg = session_context->uncompleted_message;
 				}
 				else if (IsA(node, ExecuteStmt))
-					ps = pool_get_prepared_statement_by_pstmt_name(((ExecuteStmt *) node)->name);
+				{
+					msg = pool_get_sent_message('Q', ((ExecuteStmt *)node)->name);
+					if (!msg)
+						msg = pool_get_sent_message('P', ((ExecuteStmt *)node)->name);
+				}
 
 				/* rewrite `now()' to timestamp literal */
-				rewrite_query = rewrite_timestamp(backend, query_context->parse_tree, false, ps);
+				rewrite_query = rewrite_timestamp(backend, query_context->parse_tree, false, msg);
 				if (rewrite_query != NULL)
 				{
 					query_context->rewritten_query = rewrite_query;
@@ -396,16 +430,14 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			{
 				/* Send query to all DB nodes at once */
 				status = pool_send_and_wait(query_context, string, len, 0, 0, "");
-				/*
-				free_parser();
-				*/
+				/* free_parser(); */
 				return status;
 			}
 
 			/* Send the query to master node */
 			if (pool_send_and_wait(query_context, string, len, 1, MASTER_NODE_ID, "") != POOL_CONTINUE)
 			{
-				free_parser();
+				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
 		}
@@ -415,7 +447,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		 */
 		if (pool_send_and_wait(query_context, string, len, -1, MASTER_NODE_ID, "") != POOL_CONTINUE)
 		{
-			free_parser();
+			pool_query_context_destroy(query_context);
 			return POOL_END;
 		}
 
@@ -424,25 +456,25 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		{
 			if (pool_send_and_wait(query_context, string, len, 1, MASTER_NODE_ID, "") != POOL_CONTINUE)
 			{
-/*
-				free_parser();
-*/
+				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
 		}
-		free_parser();
+		/* free_parser(); */
 	}
 	else
 	{
 		if (pool_send_and_wait(query_context, string, len, 1, MASTER_NODE_ID, "") != POOL_CONTINUE)
 		{
-			free_parser();
+			pool_query_context_destroy(query_context);
 			return POOL_END;
 		}
-/*
-		free_parser();
-*/
+		/* free_parser(); */
 	}
+
+	/* switch memory context */
+	pool_memory = old_context;
+
 	return POOL_CONTINUE;
 }
 
@@ -453,12 +485,12 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 					int len, char *contents)
 {
 	int commit = 0;
-	Portal *portal;
 	char *query = NULL;
 	Node *node;
 	int specific_error = 0;
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
+	POOL_SENT_MESSAGE *msg;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -470,31 +502,31 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 	pool_debug("Execute: portal name <%s>", contents);
 
-	portal = pool_get_portal_by_portal_name(contents);
-	if (portal == NULL)
+	msg = pool_get_sent_message('B', contents);
+	if (!msg)
 	{
-		pool_error("Execute: cannot get portal");
+		pool_error("Execute: cannot get bind message");
 		return POOL_END;
 	}
-	if (portal->pstmt == NULL)
-	{
-		pool_error("Execute: cannot get prepared statement");
-		return POOL_END;
-	}
-	if (portal->pstmt->qctxt == NULL)
+	if(!msg->query_context)
 	{
 		pool_error("Execute: cannot get query context");
 		return POOL_END;
 	}
-	if (portal->pstmt->qctxt->parse_tree== NULL)
+	if (!msg->query_context->original_query)
+	{
+		pool_error("Execute: cannot get original query");
+		return POOL_END;
+	}
+	if (!msg->query_context->parse_tree)
 	{
 		pool_error("Execute: cannot get parse tree");
 		return POOL_END;
 	}
 
-	query_context = portal->pstmt->qctxt;
-	node = query_context->parse_tree;
-	query = portal->pstmt->qctxt->original_query;
+	query_context = msg->query_context;
+	node = msg->query_context->parse_tree;
+	query = msg->query_context->original_query;
 	pool_debug("Execute: query: %s", query);
 	strncpy(query_string_buffer, query, sizeof(query_string_buffer));
 
@@ -504,10 +536,6 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	session_context->query_context = query_context;
 	pool_where_to_send(query_context, query, node);
 
-
-	if (IsA(query_context->parse_tree, DeallocateStmt))
-		overwrite_map_for_deallocate(query_context);
-
 	/* check if query is "COMMIT" or "ROLLBACK" */
 	commit = is_commit_query(node);
 
@@ -587,7 +615,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	char *stmt;
 	List *parse_tree_list;
 	Node *node = NULL;
-	PreparedStatement *ps;
+	POOL_SENT_MESSAGE *msg;
 	POOL_STATUS status;
 	POOL_MEMORY_POOL *old_context;
 	POOL_SESSION_CONTEXT *session_context;
@@ -618,17 +646,31 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 	if (parse_tree_list == NIL)
 	{
-		/*
-		 * Unable to parse the query. Probably syntax error or the
-		 * query is too new and our parser cannot understand. Treat as
-		 * if it were an DELETE command. Note that the DELETE command
-		 * does not execute, instead the original query will be sent
-		 * to backends, which may or may not cause an actual syntax errors.
-		 * The command will be sent to all backends in replication mode
-		 * or master/primary in master/slave mode.
-		 */
-		pool_log("Parse: Unable to parse the query: %s", contents);
-		parse_tree_list = raw_parser(POOL_DUMMY_QUERY);
+		/* is the query empty? */
+		if (*stmt == '\0' || *stmt == ';')
+		{
+			/*
+			 * JBoss sends empty queries for checking connections.
+			 * We replace the empty query with SELECT command not
+			 * to affect load balance.
+			 * [Pgpool-general] Confused about JDBC and load balancing
+			 */
+			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY);
+		}
+		else
+		{
+			/*
+			 * Unable to parse the query. Probably syntax error or the
+			 * query is too new and our parser cannot understand. Treat as
+			 * if it were an DELETE command. Note that the DELETE command
+			 * does not execute, instead the original query will be sent
+			 * to backends, which may or may not cause an actual syntax errors.
+			 * The command will be sent to all backends in replication mode
+			 * or master/primary in master/slave mode.
+			 */
+			pool_log("Parse: Unable to parse the query: %s", stmt);
+			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY);
+		}
 	}
 
 	if (parse_tree_list != NIL)
@@ -645,8 +687,14 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		 */
 		pool_start_query(query_context, pstrdup(stmt), node);
 
-		ps = pool_create_prepared_statement(name, 0, len, contents, query_context);
-		session_context->pending_pstmt = ps;
+		msg = pool_create_sent_message('P', len, contents, 0, name, query_context);
+		if (!msg)
+		{
+			pool_error("Parse: cannot create parse message: %s", strerror(errno));
+			return POOL_END;
+		}
+
+		session_context->uncompleted_message = msg;
 
 		/*
 		 * Decide where to send query
@@ -654,9 +702,6 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		pool_where_to_send(query_context, query_context->original_query,
 						   query_context->parse_tree);
 
-		if (IsA(query_context->parse_tree, DeallocateStmt))
-			overwrite_map_for_deallocate(query_context);
-
 		if (REPLICATION)
 		{
 			char *rewrite_query;
@@ -670,8 +715,8 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			 */
 			if (*name == '\0')
 				rewrite_to_params = false;
-			ps->num_tsparams = 0;
-			rewrite_query = rewrite_timestamp(backend, node, rewrite_to_params, ps);
+			msg->num_tsparams = 0;
+			rewrite_query = rewrite_timestamp(backend, node, rewrite_to_params, msg);
 			if (rewrite_query != NULL)
 			{
 				int alloc_len = len - strlen(stmt) + strlen(rewrite_query);
@@ -687,8 +732,10 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				stmt = contents + strlen(name) + 1;
 				pool_debug("Parse: rewrite query  %s %s len=%d", name, stmt, len);
 
-				ps->parse_len = len;
-				ps->parse_contents = contents;
+				msg->len = len;
+				msg->contents = contents;
+
+				query_context->rewritten_query = rewrite_query;
 			}
 		}
 	}
@@ -715,15 +762,22 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			kind = pool_read_kind(backend);
 			if (kind != 'Z')
 			{
-				/* free_parser(); */
+				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
 
 			if (ReadyForQuery(frontend, backend, 0) != POOL_CONTINUE)
 			{
-				/* free_parser(); */
+				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
+
+			/*
+			 * set in_progress flag, because ReadyForQuery unset it.
+			 * in_progress flag influences VALID_BACKEND.
+			 */
+			if (!pool_is_query_in_progress())
+				pool_set_query_in_progress();
 		}
 
 		if (is_strict_query(query_context->parse_tree))
@@ -735,7 +789,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			status = insert_lock(frontend, backend, stmt, (InsertStmt *)query_context->parse_tree, insert_stmt_with_lock);
 			if (status != POOL_CONTINUE)
 			{
-				/* free_parser(); */
+				pool_query_context_destroy(query_context);
 				return status;
 			}
 		}
@@ -755,7 +809,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		pool_debug("Parse: waiting for master completing the query");
 		if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "P") != POOL_CONTINUE)
 		{
-			/* free_parser(); */
+			pool_query_context_destroy(query_context);
 			return POOL_END;
 		}
 
@@ -768,7 +822,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
 		if (deadlock_detected < 0)
 		{
-			/* free_parser(); */
+			pool_query_context_destroy(query_context);
 			return POOL_END;
 		}
 		else
@@ -789,7 +843,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 								   strlen(POOL_ERROR_QUERY)+1, -1,
 								   MASTER_NODE_ID, "") != POOL_CONTINUE)
 			{
-				/* free_parser(); */
+				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
 		}
@@ -797,7 +851,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		{
 			if (pool_send_and_wait(query_context, contents, len, -1, MASTER_NODE_ID, "P") != POOL_CONTINUE)
 			{
-				/* free_parser(); */
+				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
 		}
@@ -806,7 +860,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	{
 		if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "P") != POOL_CONTINUE)
 		{
-			/* free_parser(); */
+			pool_query_context_destroy(query_context);
 			return POOL_END;
 		}
 	}
@@ -826,8 +880,8 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	char *pstmt_name;
 	char *portal_name;
 	char *rewrite_msg;
-	Portal *portal = NULL;
-	PreparedStatement *pstmt = NULL;
+	POOL_SENT_MESSAGE *parse_msg;
+	POOL_SENT_MESSAGE *bind_msg;
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
@@ -845,33 +899,37 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	portal_name = contents;
 	pstmt_name = contents + strlen(portal_name) + 1;
 
-	pstmt = pool_get_prepared_statement_by_pstmt_name(pstmt_name);
-	if (pstmt == NULL)
+	parse_msg = pool_get_sent_message('Q', pstmt_name);
+	if (!parse_msg)
+		parse_msg = pool_get_sent_message('P', pstmt_name);
+	if (!parse_msg)
 	{
-		pool_error("Bind: cannot get prepared statement \"%s\"", pstmt_name);
+		pool_error("Bind: cannot get parse message \"%s\"", pstmt_name);
 		return POOL_END;
 	}
 
-	portal = pool_create_portal(portal_name, pstmt->num_tsparams, pstmt);
-	if (portal == NULL)
+	bind_msg = pool_create_sent_message('B', len, contents,
+										parse_msg->num_tsparams, portal_name,
+										parse_msg->query_context);
+	if (!bind_msg)
 	{
-		pool_error("Bind: cannot create portal: %s", strerror(errno));
+		pool_error("Bind: cannot create bind message: %s", strerror(errno));
 		return POOL_END;
 	}
 
-	query_context = pstmt->qctxt;
-	if (query_context == NULL)
+	query_context = parse_msg->query_context;
+	if (!query_context)
 	{
 		pool_error("Bind: cannot get query context");
 		return POOL_END;
 	}
 
-	session_context->pending_portal = portal;
+	session_context->uncompleted_message = bind_msg;
 
 	/* rewrite bind message */
-	if (REPLICATION && portal->num_tsparams > 0)
+	if (REPLICATION && bind_msg->num_tsparams > 0)
 	{
-		rewrite_msg = bind_rewrite_timestamp(backend, portal, contents, &len);
+		rewrite_msg = bind_rewrite_timestamp(backend, bind_msg, contents, &len);
 		if (rewrite_msg != NULL)
 			contents = rewrite_msg;
 	}
@@ -880,12 +938,9 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	pool_where_to_send(query_context, query_context->original_query,
 					   query_context->parse_tree);
 
-	if (IsA(query_context->parse_tree, DeallocateStmt))
-		overwrite_map_for_deallocate(query_context);
-
 	if (pool_config->load_balance_mode && pool_is_writing_transaction())
 	{
-		if(parse_before_bind(frontend, backend, pstmt) != POOL_CONTINUE)
+		if(parse_before_bind(frontend, backend, parse_msg) != POOL_CONTINUE)
 			return POOL_END;
 	}
 
@@ -915,8 +970,7 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 							int len, char *contents)
 {
-	Portal *portal = NULL;
-	PreparedStatement *pstmt = NULL;
+	POOL_SENT_MESSAGE *msg;
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
@@ -931,28 +985,28 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	/* Prepared Statement */
 	if (*contents == 'S')
 	{
-		pstmt = pool_get_prepared_statement_by_pstmt_name(contents+1);
+		msg = pool_get_sent_message('Q', contents+1);
+		if (!msg)
+			msg = pool_get_sent_message('P', contents+1);
+		if (!msg)
+		{
+			pool_error("Describe: cannot get parse message");
+			return POOL_END;
+		}
 	}
 	/* Portal */
 	else
 	{
-		portal = pool_get_portal_by_portal_name(contents+1);
-		if (portal == NULL)
+		msg = pool_get_sent_message('B', contents+1);
+		if (!msg)
 		{
-			pool_error("Describe: cannot get portal \"%s\"", contents+1);
+			pool_error("Describe: cannot get bind message");
 			return POOL_END;
 		}
-
-		pstmt = portal->pstmt;
 	}
 
-	if (pstmt == NULL)
-	{
-		pool_error("Describe: cannot get prepared statement");
-		return POOL_END;
-	}
+	query_context = msg->query_context;
 
-	query_context = pstmt->qctxt;
 	if (query_context == NULL)
 	{
 		pool_error("Describe: cannot get query context");
@@ -963,9 +1017,6 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	pool_where_to_send(query_context, query_context->original_query,
 					   query_context->parse_tree);
 
-	if (IsA(query_context->parse_tree, DeallocateStmt))
-		overwrite_map_for_deallocate(query_context);
-
 	pool_debug("Describe: waiting for master completing the query");
 	if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "D")
 		!= POOL_CONTINUE)
@@ -982,8 +1033,7 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				  int len, char *contents)
 {
-	Portal *portal = NULL;
-	PreparedStatement *pstmt = NULL;
+	POOL_SENT_MESSAGE *msg;
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
@@ -998,28 +1048,24 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	/* Prepared Statement */
 	if (*contents == 'S')
 	{
-		pstmt = pool_get_prepared_statement_by_pstmt_name(contents+1);
-		if (pstmt == NULL)
+		msg = pool_get_sent_message('Q', contents+1);
+		if (!msg)
+			msg = pool_get_sent_message('P', contents+1);
+		if (!msg)
 		{
-			pool_error("Close: cannot get prepared statement");
+			pool_error("Close: cannot get parse message");
 			return POOL_END;
 		}
-
-		session_context->pending_pstmt = pstmt;
-		query_context = pstmt->qctxt;
 	}
 	/* Portal */
 	else if (*contents == 'P')
 	{
-		portal = pool_get_portal_by_portal_name(contents+1);
-		if (portal == NULL)
+		msg = pool_get_sent_message('B', contents+1);
+		if (!msg)
 		{
-			pool_error("Close: cannot get portal");
+			pool_error("Close: cannot get bind message");
 			return POOL_END;
 		}
-
-		session_context->pending_portal = portal;
-		query_context = portal->qctxt;
 	}
 	else
 	{
@@ -1027,7 +1073,10 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		return POOL_END;
 	}
 
-	if (query_context == NULL)
+	session_context->uncompleted_message = msg;
+	query_context = msg->query_context;
+
+	if (!query_context)
 	{
 		pool_error("Close: cannot get query context");
 		return POOL_END;
@@ -1086,6 +1135,8 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	signed char kind;
 	signed char state;
 	POOL_SESSION_CONTEXT *session_context;
+	Node *node = NULL;
+	char *query = NULL;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -1157,6 +1208,10 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				degenerate_backend_set(victim_nodes, number_of_nodes);
 				child_exit(1);
 			}
+			else
+			{
+				pool_error("ReadyForQuery: find_victim_nodes returned no victim node");
+			}
 		}
 
 		/*
@@ -1226,7 +1281,8 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		/*
 		 * Set transaction state for each node
 		 */
-		state = TSTATE(backend, MASTER_NODE_ID);
+		state = TSTATE(backend,
+					   MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID);
 
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
@@ -1242,7 +1298,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 			 * The transaction state to be returned to frontend is
 			 * master's.
 			 */
-			if (i == MASTER_NODE_ID)
+			if (i == (MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID))
 			{
 				state = kind;
 			}
@@ -1264,9 +1320,6 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 
 	if (pool_is_query_in_progress() && pool_is_command_success())
 	{
-		Node *node;
-		char *query;
-
 		node = pool_get_parse_tree();
 		query = pool_get_query_string();
 
@@ -1298,14 +1351,14 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 			 * If 2PC commands, automatically close transaction on standbys since
 			 * 2PC commands close transaction on primary.
 			 */
-			else if (is_2pc_transaction_query(node, query))
+			else if (is_2pc_transaction_query(node))
 			{
 				for (i=0;i<NUM_BACKENDS;i++)
 				{
 					if (CONNECTION_SLOT(backend, i) &&
 						TSTATE(backend, i) == 'T' &&
 						BACKEND_INFO(i).backend_status == CON_UP &&
-						REAL_MASTER_NODE_ID != i)
+						(MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) != i)
 					{
 						per_node_statement_log(backend, i, "COMMIT");
 						if (do_command(frontend, CONNECTION(backend, i), "COMMIT", MAJOR(backend), 
@@ -1331,7 +1384,10 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	}
 
 	if (!pool_is_doing_extended_query_message())
-		pool_query_context_destroy(pool_get_session_context()->query_context);
+	{
+		if (!(node && IsA(node, PrepareStmt)))
+			pool_query_context_destroy(pool_get_session_context()->query_context);
+	}
 
 	sp = MASTER_CONNECTION(backend)->sp;
 	if (MASTER(backend)->tstate == 'T')
@@ -1357,18 +1413,17 @@ POOL_STATUS ParseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 		return POOL_END;
 	}
 
-	if (session_context->pending_pstmt != NULL)
+	if (session_context->uncompleted_message)
 	{
 		POOL_QUERY_CONTEXT *qc;
 
-		pool_add_prepared_statement();
+		pool_add_sent_message(session_context->uncompleted_message);
 
-		/* Set "parse done" to query_state */
-		qc = session_context->pending_pstmt->qctxt;
-		if (qc != NULL)
-			pool_set_query_state(qc, 1);
+		qc = session_context->uncompleted_message->query_context;
+		if (qc)
+			pool_set_query_state(qc, POOL_PARSE_COMPLETE);
 
-		session_context->pending_pstmt = NULL;
+		session_context->uncompleted_message = NULL;
 	}
 
 	return SimpleForwardToFrontend('1', frontend, backend);
@@ -1386,18 +1441,17 @@ POOL_STATUS BindComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backen
 		return POOL_END;
 	}
 
-	if (session_context->pending_portal != NULL)
+	if (session_context->uncompleted_message)
 	{
-		PreparedStatement *pstmt;
+		POOL_QUERY_CONTEXT *qc;
 
-		pool_add_portal();
+		pool_add_sent_message(session_context->uncompleted_message);
 
-		/* Set "bind done" to query_state */
-		pstmt = session_context->pending_portal->pstmt;
-		if (pstmt != NULL && pstmt->qctxt != NULL)
-			pool_set_query_state(pstmt->qctxt, 2);
+		qc = session_context->uncompleted_message->query_context;
+		if (qc)
+			pool_set_query_state(qc, POOL_BIND_COMPLETE);
 
-		session_context->pending_portal = NULL;
+		session_context->uncompleted_message = NULL;
 	}
 
 	return SimpleForwardToFrontend('2', frontend, backend);
@@ -1406,6 +1460,7 @@ POOL_STATUS BindComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backen
 POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
 {
 	POOL_SESSION_CONTEXT *session_context;
+	POOL_STATUS status;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -1415,23 +1470,23 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 		return POOL_END;
 	}
 
-	if (session_context->pending_pstmt != NULL)
+	/* Send CloseComplete(3) to frontend before removing the target message */
+	status = SimpleForwardToFrontend('3', frontend, backend);
+
+	/* Remove the target message */
+	if (session_context->uncompleted_message)
 	{
-		pool_remove_prepared_statement();
-		session_context->pending_pstmt = NULL;
-	}
-	else if (session_context->pending_portal != NULL)
-	{
-		pool_remove_portal();
-		session_context->pending_portal = NULL;
+		pool_remove_sent_message(session_context->uncompleted_message->kind,
+								 session_context->uncompleted_message->name);
+		session_context->uncompleted_message = NULL;
 	}
 	else
 	{
-		pool_error("CloseComplete: pending object not found");
+		pool_error("CloseComplete: uncompleted message not found");
 		return POOL_END;
 	}
 
-	return SimpleForwardToFrontend('3', frontend, backend);
+	return status;
 }
 
 POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
@@ -1458,8 +1513,11 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 
 		if (IsA(node, PrepareStmt))
 		{
-			pool_add_prepared_statement();
-			session_context->pending_pstmt = NULL;
+			if (session_context->uncompleted_message)
+			{
+				pool_add_sent_message(session_context->uncompleted_message);
+				session_context->uncompleted_message = NULL;
+			}
 		}
 		else if (IsA(node, DeallocateStmt))
 		{
@@ -1467,18 +1525,28 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 			
 			name = ((DeallocateStmt *)node)->name;
 			if (name == NULL)
-				pool_clear_prepared_statement_list();
+			{
+				pool_remove_sent_messages('Q');
+				pool_remove_sent_messages('P');
+			}
 			else
-				pool_remove_prepared_statement();
-			session_context->pending_pstmt = NULL;
+			{
+				pool_remove_sent_message('Q', name);
+				pool_remove_sent_message('P', name);
+			}
 		}
 		else if (IsA(node, DiscardStmt))
 		{
 			DiscardStmt *stmt = (DiscardStmt *)node;
-			if (stmt->target == DISCARD_ALL || stmt->target == DISCARD_PLANS)
+
+			if (stmt->target == DISCARD_PLANS)
 			{
-				pool_remove_pending_objects();
-				pool_clear_prepared_statement_list();
+				pool_remove_sent_messages('Q');
+				pool_remove_sent_messages('P');
+			}
+			else if (stmt->target == DISCARD_ALL)
+			{
+				pool_clear_sent_message_list();
 			}
 		}
 		/*
@@ -1678,6 +1746,7 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 						   POOL_CONNECTION_POOL *backend)
 {
 	POOL_STATUS ret;
+	POOL_SESSION_CONTEXT *session_context;
 
 	ret = SimpleForwardToFrontend('E', frontend, backend);
 	if (ret != POOL_CONTINUE)
@@ -1745,7 +1814,17 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 	/* An error occurred with PREPARE or DEALLOCATE command.
 	 * Free pending portal object.
 	 */
-	pool_remove_pending_objects();
+	session_context = pool_get_session_context();
+	if (session_context)
+	{
+		if (session_context->uncompleted_message)
+		{
+			pool_add_sent_message(session_context->uncompleted_message);
+			pool_remove_sent_message(session_context->uncompleted_message->kind,
+									 session_context->uncompleted_message->name);
+			session_context->uncompleted_message = NULL;
+		}
+	}
 
 	return POOL_CONTINUE;
 }
@@ -1950,7 +2029,6 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 
 		case 'S':  /* Sync */
 			pool_set_doing_extended_query_message();
-			receive_extended_begin = 0;
 			if (pool_is_ignore_till_sync())
 				pool_unset_ignore_till_sync();
 			if (!pool_is_query_in_progress())
@@ -2079,7 +2157,9 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 			case 'E':	/* ErrorResponse */
 				status = ErrorResponse3(frontend, backend);
-				if (TSTATE(backend, REAL_MASTER_NODE_ID) != 'I')
+				pool_set_command_success();
+				if (TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID :
+						   REAL_MASTER_NODE_ID) != 'I')
 					pool_set_failed_transaction();
 				if (pool_is_doing_extended_query_message())
 				{
@@ -2148,7 +2228,8 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 			case 'E':	/* ErrorResponse */
 				status = ErrorResponse(frontend, backend);
-				if (TSTATE(backend, REAL_MASTER_NODE_ID) != 'I')
+				if (TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID :
+						   REAL_MASTER_NODE_ID) != 'I')
 					pool_set_failed_transaction();
 				break;
 
@@ -2481,21 +2562,21 @@ POOL_STATUS raise_intentional_error_if_need(POOL_CONNECTION_POOL *backend)
 	query_context = session_context->query_context;
 
 	if (MASTER_SLAVE &&
-		TSTATE(backend, REAL_MASTER_NODE_ID) == 'T' &&
-		REAL_MASTER_NODE_ID != MASTER_NODE_ID &&
+		TSTATE(backend, PRIMARY_NODE_ID) == 'T' &&
+		PRIMARY_NODE_ID != MASTER_NODE_ID &&
 		query_context &&
 		is_select_query(query_context->parse_tree, query_context->original_query))
 	{
-		pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+		pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 		if (pool_is_doing_extended_query_message())
 		{
-			ret = do_error_execute_command(backend, REAL_MASTER_NODE_ID, PROTO_MAJOR_V3);
+			ret = do_error_execute_command(backend, PRIMARY_NODE_ID, PROTO_MAJOR_V3);
 			if (ret != POOL_CONTINUE)
 				return ret;
 		}
 		else
 		{
-			ret = do_error_command(CONNECTION(backend, REAL_MASTER_NODE_ID), MAJOR(backend));
+			ret = do_error_command(CONNECTION(backend, PRIMARY_NODE_ID), MAJOR(backend));
 			if (ret != POOL_CONTINUE)
 				return ret;
 		}
@@ -2710,23 +2791,23 @@ void per_node_error_log(POOL_CONNECTION_POOL *backend, int node_id, char *query,
 
 static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 									 POOL_CONNECTION_POOL *backend,
-									 PreparedStatement *ps)
+									 POOL_SENT_MESSAGE *message)
 {
 	int i;
-	int len = ps->parse_len;
+	int len = message->len;
 	char kind = '\0';
-	char *contents = ps->parse_contents;
+	char *contents = message->contents;
 	bool parse_was_sent = false;
 	bool backup[MAX_NUM_BACKENDS];
 	POOL_STATUS status;
-	POOL_QUERY_CONTEXT *qc = ps->qctxt;
+	POOL_QUERY_CONTEXT *qc = message->query_context;
 
 	memcpy(backup, qc->where_to_send, sizeof(qc->where_to_send));
 
 	/* expect to send to master node only */
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
-		if (qc->where_to_send[i] && qc->query_state[i] < 1)	/* 1: parse done */
+		if (qc->where_to_send[i] && statecmp(qc->query_state[i], POOL_PARSE_COMPLETE) < 0)
 		{
 			pool_debug("parse_before_bind: waiting for backend %d completing parse", i);
 			if (pool_send_and_wait(qc, contents, len, 1, i, "P") != POOL_CONTINUE)
@@ -2770,30 +2851,6 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
-static void overwrite_map_for_deallocate(POOL_QUERY_CONTEXT *query_context)
-{
-	char *name;
-	PreparedStatement *ps;
-	POOL_QUERY_CONTEXT *qc;
-
-	if (IsA(query_context->parse_tree, DeallocateStmt)) {
-		name = ((DeallocateStmt *)query_context->parse_tree)->name;
-		if (name != NULL)	/* NULL is "DEALLOCATE ALL" */
-		{
-			ps = pool_get_prepared_statement_by_pstmt_name(name);
-			if (ps != NULL)
-			{
-				qc = ps->qctxt;
-				if (qc != NULL)
-				{
-					memcpy(query_context->where_to_send, qc->where_to_send,
-						   sizeof(query_context->where_to_send));
-				}
-			}
-		}
-	}
-}
-
 /*
  * Find victim nodes by "decide by majority" rule and returns array
  * of victim node ids. If no victim is found, return NULL.
@@ -2804,7 +2861,7 @@ static void overwrite_map_for_deallocate(POOL_QUERY_CONTEXT *query_context)
  * master_node: The master node id. Less than 0 means ignore this parameter.
  * number_of_nodes: Number of elements in victim nodes array.
  *
- * Note: If no one wins and master_node > 0, winner would be the
+ * Note: If no one wins and master_node >= 0, winner would be the
  * master and other nodes who has same number of tuples as the master.
  *
  * Caution: Returned victim node array is allocated in static memory
@@ -2900,7 +2957,7 @@ static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *
 	/* Make victim nodes list */
 	for (i=0;i<nmembers;i++)
 	{
-		if (ntuples[i] > 0 && ntuples[i] != majority_ntuples)
+		if (ntuples[i] >= 0 && ntuples[i] != majority_ntuples)
 		{
 			victim_nodes[(*number_of_nodes)++] = i;
 		}
